#!/usr/bin/env python3

import sys
import argparse

from AaronTools.geometry import Geometry
from AaronTools.component import Component
from AaronTools.finders import AnyTransitionMetal
from AaronTools.fileIO import FileReader, read_types
from AaronTools.utils.utils import glob_files

# from cProfile import Profile

solid_parser = argparse.ArgumentParser(
    description="print ligand solid angle",
    formatter_class=argparse.RawTextHelpFormatter
)

solid_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file",
)

solid_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    required=False,
    dest="outfile",
    help="output destination\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: stdout"
)

solid_parser.add_argument(
    "-if", "--input-format",
    type=str,
    default=None,
    dest="input_format",
    choices=read_types,
    help="file format of input - xyz is assumed if input is stdin",
)

solid_parser.add_argument(
    "-k", "--key-atoms",
    type=str,
    default=None,
    dest="key_atoms",
    help="indices of ligand coordinating atoms you are calculating\n" +
    "the cone angle of (1-indexed)",
)

solid_parser.add_argument(
    "-c", "--center",
    type=str,
    default=AnyTransitionMetal(),
    dest="center",
    help="index of complex's center atom (1-indexed)\nDefault: transition metals",
)

solid_parser.add_argument(
    "-p", "--points",
    type=int,
    choices=[110, 194, 302, 590, 974, 1454, 2030, 2702, 5810],
    default=5810,
    dest="points",
    help="number of angular points for integration\n" +
    "lower values are faster, but at the cost of accuracy\nDefault: 5810"
)

solid_parser.add_argument(
    "-sca", "--solid-cone-angle",
    action="store_true",
    default=False,
    dest="solid_cone",
    help="print solid ligand cone angles instead of solid angles",
)

solid_parser.add_argument(
    "-vdw", "--vdw-radii",
    default="umn",
    choices=["umn", "bondi"],
    dest="radii",
    help="VDW radii to use in calculation\n" + 
    "umn: main group vdw radii from J. Phys. Chem. A 2009, 113, 19, 5806–5812\n" +
    "    (DOI: 10.1021/jp8111556)\n" + 
    "    transition metals are crystal radii from Batsanov, S.S. Van der Waals\n" +
    "    Radii of Elements. Inorganic Materials 37, 871–885 (2001).\n" +
    "    (DOI: 10.1023/A:1011625728803)\n" + 
    "bondi: radii from J. Phys. Chem. 1964, 68, 3, 441–451 (DOI: 10.1021/j100785a001)\n" +
    "Default: umn",
)


args = solid_parser.parse_args()

# profile = Profile()
# profile.enable()

s = ""

# args.radii = {
#     "P": 1.8,
#     "H": 1.2,
#     "C": 1.7,
#     "N": 1.55,
#     "O": 1.52,
#     "F": 1.47,
#     "S": 1.8,
#     "Cl": 1.75,
#     "Fe": 2.0,
# }

for f in glob_files(args.infile, parser=solid_parser):
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format[0], None))
        else:
            infile = FileReader(f, just_geom=False)
    else:
        if args.input_format is not None:
            infile = FileReader(("from stdin", args.input_format[0], f))
        else:
            if len(sys.argv) >= 1:
                infile = FileReader(("from stdin", "xyz", f))

    geom = Geometry(infile, refresh_ranks=False)

    ligand = geom.get_fragment(args.key_atoms, stop=args.center)

    comp = Component(
        ligand,
        key_atoms=args.key_atoms,
        detect_backbone=False,
        refresh_ranks=False,
        refresh_connected=False,
    )

    angle = comp.solid_angle(
        center=geom.find(args.center),
        radii=args.radii,
        return_solid_cone=args.solid_cone,
        grid=args.points,
    )
    
    if len(args.infile) > 1:
        s += "%20s:\t" % f

    if args.solid_cone:
        s += "%4.2f\n" % angle
    else:
        s += "%4.3f\n" % angle
    
if not args.outfile:
    print(s.rstrip())
else:
    with open(args.outfile, "a") as f:
        f.write(s.rstrip())

# profile.disable()
# profile.print_stats()


#!/usr/bin/env python3
"""basically cat_screen from the perl version"""

import argparse
import sys

from AaronTools.fileIO import FileReader, read_types
from AaronTools.geometry import Geometry
from AaronTools.substituent import Substituent
from AaronTools.utils.utils import get_filename, glob_files

substitute_parser = argparse.ArgumentParser(
    description="add or modify substituents with permutations",
    formatter_class=argparse.RawTextHelpFormatter,
)
substitute_parser.add_argument(
    "infile",
    metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file",
)

substitute_parser.add_argument(
    "-ls",
    "--list",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="list_avail",
    help="list available substituents",
)

substitute_parser.add_argument(
    "-if",
    "--input-format",
    type=str,
    default=None,
    choices=read_types,
    dest="input_format",
    help="file format of input - xyz is assumed if input is stdin",
)

substitute_parser.add_argument(
    "-s",
    "--substitute",
    metavar="i,j,k,...=substituent1,substituent2,...",
    type=str,
    action="append",
    default=None,
    required=False,
    dest="substitutions",
    help="substitution instructions \n"
    + "i, j, k are the 1-indexed position of the starting position of the\n"
    + "substituent(s) you are replacing\n"
    + "a substituent name prefixed by iupac: or smiles: (e.g. iupac:acetyl\n"
    + "or smiles:O=[N.]=O) will create the substituent from the\n"
    + "corresponding identifier\n"
    + "substituents can be separated by commas to use all combinations",
)

substitute_parser.add_argument(
    "-m",
    "--minimize",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="mini",
    help="rotate substituents to try to minimize LJ energy",
)

substitute_parser.add_argument(
    "-o",
    "--output",
    type=str,
    default=False,
    required=False,
    metavar="output destination",
    dest="outfile",
    help="output destination\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "$SUBSTITUTIONS will be replaced with the substitution info\n" +
    "Default: stdout"
)

args = substitute_parser.parse_args()

if args.list_avail:
    sub_list = ""
    for i, name in enumerate(sorted(Substituent.list())):
        sub_list += "%-20s" % name
        # if (i + 1) % 3 == 0:
        if (i + 1) % 1 == 0:
            sub_list += "\n"

    print(sub_list.strip())
    sys.exit(0)

for infile in glob_files(args.infile, parser=substitute_parser):
    if isinstance(infile, str):
        if args.input_format is not None:
            f = FileReader((infile, args.input_format, infile))
        else:
            f = FileReader(infile)
    else:
        if args.input_format is not None:
            f = FileReader(("from stdin", args.input_format, infile))
        else:
            f = FileReader(("from stdin", "xyz", infile))

    geom = Geometry(f)

    target_list = []
    sub_names = []
    # number of geometries is the product of the number of substituents
    # that are to be placed at each site
    total_geoms = 1
    # mod_array is used to determine which substituent combination to
    # use for the ith structure
    mod_array = []
    for sub in args.substitutions:
        ndx_targets = sub.split("=")[0]
        target_list.append([a.name for a in geom.find(ndx_targets)])
        sub_name = "=".join(sub.split("=")[1:])
        sub_names.append([x.strip() for x in sub_name.split(",")])
        total_geoms *= len(sub_names[-1])
        mod_array = [len(sub_names[-1]) * x for x in mod_array]
        mod_array.append(1)

    for i in range(0, total_geoms):
        new_geom = geom.copy()

        sub_str = ""

        # go through each substitution request combination
        for j, (ndx, sub_list) in enumerate(zip(target_list, sub_names)):
            sub_ndx = int(i / mod_array[j]) % len(sub_list)
            sub_name = sub_list[sub_ndx]
            sub_str += "_%s=%s" % (",".join(ndx), sub_name)
            for target in ndx:
                if sub_name.lower().startswith("iupac:"):
                    sub_name = ":".join(sub_name.split(":")[1:])
                    sub = Substituent.from_string(sub_name, form="iupac")
                elif sub_name.lower().startswith("smiles:"):
                    sub_name = ":".join(sub_name.split(":")[1:])
                    sub = Substituent.from_string(sub_name, form="smiles")
                else:
                    sub = Substituent(sub_name)

                # replace old substituent with new substituent
                new_geom.substitute(sub, target, minimize=args.mini)
                new_geom.refresh_connected()

        sub_str = sub_str.strip("_")

        if args.outfile:
            outfile = args.outfile
            # only append if the file name is probably not unique
            do_append = True
            if "$INFILE" in outfile:
                outfile = outfile.replace("$INFILE", get_filename(infile))
                do_append = False
            if "$SUBSTITUTIONS" in outfile:
                outfile = outfile.replace("$SUBSTITUTIONS", sub_str)
                do_append = False
            new_geom.write(
                append=do_append,
                outfile=outfile,
            )
        else:
            print(new_geom.write(outfile=False))



import collections.abc
import os
import re
from collections import OrderedDict
from math import acos, sin, cos, sqrt

import AaronTools.atoms as Atoms
import numpy as np
from AaronTools.const import AARONTOOLS, PHYSICAL

_loaded_leb_grids = dict()


def range_list(number_list, sep=",", sort=True):
    """
    Takes a list of numbers and puts them into a string containing ranges, eg:
    [1, 2, 3, 5, 6, 7, 9, 10] -> "1-3,5-7,9,10"

    :sep: the separator to use between consecutive ranges
    :sort: sort the list before parsing
    """
    if sort:
        number_list = sorted(number_list)
    tmp = [[]]
    for i, n in enumerate(number_list):
        if i == 0:
            tmp[-1] += [n]
        elif n == number_list[i - 1] + 1:
            tmp[-1] += [n]
        else:
            tmp += [[n]]
    rv = ""
    for t in tmp:
        if len(t) > 2:
            rv += "{}-{},".format(t[0], t[-1])
        elif len(t) == 2:
            rv += "{},{},".format(t[0], t[-1])
        else:
            rv += "{},".format(t[0])
    return rv[:-1]


def progress_bar(this, max_num, name=None, width=50):
    if name is None:
        out_str = ""
    else:
        out_str = "{}: ".format(name)
    out_str += "Progress {:3.0f}% ".format(100 * this / max_num)
    out_str += "|{:<{width}s}|".format(
        "#" * int(width * this / max_num), width=width
    )
    print(out_str, end="\r")


def clean_progress_bar(width=50):
    print(" " * 2 * width, end="\r")


def proj(v_vec, u_vec):
    """
    projection of u_vec into v_vec
    v_vec should be a np.array, and u_vec should have the
    same shape as v_vec"""
    numerator = np.dot(u_vec, v_vec)
    denominator = np.dot(v_vec, v_vec)
    return numerator * v_vec / denominator


def quat_matrix(pt1, pt2):
    """build quaternion matrix from pt1 and pt2"""
    pt1 = np.array(pt1, dtype=np.longdouble)
    pt2 = np.array(pt2, dtype=np.longdouble)
    for pt in [pt1, pt2]:
        if len(pt.shape) != 1 or pt.shape[0] != 3:
            raise ValueError("Arguments should be 3-element vectors")

    xm, ym, zm = tuple(pt1 - pt2)
    xp, yp, zp = tuple(pt1 + pt2)

    matrix = np.array(
        [
            [
                xm * xm + ym * ym + zm * zm,
                yp * zm - ym * zp,
                xm * zp - xp * zm,
                xp * ym - xm * yp,
            ],
            [
                yp * zm - ym * zp,
                yp * yp + zp * zp + xm * xm,
                xm * ym - xp * yp,
                xm * zm - xp * zp,
            ],
            [
                xm * zp - xp * zm,
                xm * ym - xp * yp,
                xp * xp + zp * zp + ym * ym,
                ym * zm - yp * zp,
            ],
            [
                xp * ym - xm * yp,
                xm * zm - xp * zp,
                ym * zm - yp * zp,
                xp * xp + yp * yp + zm * zm,
            ],
        ]
    )
    return matrix.astype(np.double)


def uptri2sym(vec, n=None, col_based=False):
    """
    Converts upper triangular matrix to a symmetric matrix

    :vec: the upper triangle array/matrix
    :n: the number of rows/columns
    :col_based: if true, triangular matirx is of the form
                    0 1 3
                    - 2 4
                    - - 5
                if false, triangular matrix is of the form
                    0 1 2
                    - 3 4
                    - - 5
    """
    if hasattr(vec[0], "__iter__") and not isinstance(vec[0], str):
        tmp = []
        for v in vec:
            tmp += v
        vec = tmp
    if n is None:
        n = -1 + np.sqrt(1 + 8 * len(vec))
        n = int(n / 2)
    if n * (n + 1) / 2 != len(vec):
        raise RuntimeError("Bad number of rows requested")

    matrix = np.zeros((n, n))
    if col_based:
        i = 0  # vector index
        j = 0  # for column counting
        for c in range(n):
            j += 1
            for r in range(n):
                matrix[r, c] = vec[i]
                matrix[c, r] = vec[i]
                i += 1
                if r + 1 == j:
                    break
    else:
        for r in range(n):
            for c in range(r, n):
                i = n * r + c - r * (1 + r) / 2
                matrix[r, c] = vec[int(i)]
                matrix[c, r] = vec[int(i)]

    return matrix


def float_vec(word):
    """
    Turns strings into floating point vectors
    :word: a comma-delimited string of numbers

    if no comma or only one element:
        returns just the floating point number
    if elements in word are strings:
        returns word unchanged
    else:
        returns a np.array() of floating point numbers
    """
    val = word
    val = val.split(",")
    try:
        val = [float(v) for v in val]
    except ValueError:
        val = word
    if len(val) == 1:
        return val[0]
    return np.array(val)


def is_alpha(test):
    """determine if string contains only letters"""
    return test.isalpha()


def is_int(test):
    rv = re.search(r"^[+-]?\d+$", test)
    return bool(rv)


def is_num(test):
    """determine if string is valid as a number"""
    rv = re.search(r"^[+-]?\d+\.?\d*", test)
    return bool(rv)


def add_dict(this, other, skip=None):
    if skip is None:
        skip = []
    for key, val in other.items():
        if key in skip:
            continue
        if key in this and isinstance(val, dict):
            add_dict(this[key], val)
        else:
            this[key] = val
    return this


def resolve_concatenation(*args):
    seq = [isinstance(a, collections.abc.MutableSequence) for a in args]
    if any(seq) and not all(seq):
        rv = []
        for i, s in enumerate(seq):
            if s:
                rv.extend(args[i])
            else:
                rv.append(args[i])
        return rv

    err_msg = "Cannot concatenate" + " {}" * len(args)
    raise TypeError(err_msg.format(*[type(a) for a in args]))


def combine_dicts(*args, case_sensitive=False, dict2_conditional=False):
    """combine dictionaries d1 and d2 to return a dictionary
    with keys d1.keys() + d2.keys()
    if a key is in d1 and d2, the items will be combined:
        if they are both dictionaries, combine_dicts is called recursively
        otherwise, d2[key] is appended to d1[key]
        if case_sensitive=False, the key in the output will be the lowercase
        of the d1 key and d2 key (only for combined items)
    dict2_conditional: bool - if True, don't add d2 keys unless they are also in d1
    """
    from copy import deepcopy

    d1 = args[0]
    d2 = args[1:]
    if len(d2) > 1:
        d2 = combine_dicts(
            d2[0],
            *d2[1:],
            case_sensitive=case_sensitive,
            dict2_conditional=dict2_conditional
        )
    else:
        d2 = d2[0]

    out = OrderedDict()
    case_keys_1 = list(d1.keys())
    case_keys_2 = list(d2.keys())
    if case_sensitive:
        keys_1 = case_keys_1
        keys_2 = case_keys_2
    else:
        keys_1 = [
            key.lower() if isinstance(key, str) else key for key in case_keys_1
        ]
        keys_2 = [
            key.lower() if isinstance(key, str) else key for key in case_keys_2
        ]

    # go through keys from d1
    for case_key, key in zip(case_keys_1, keys_1):
        # if the key is only in d1, add the item to out
        if key in keys_1 and key not in keys_2:
            out[case_key] = deepcopy(d1[case_key])
        # if the key is in both, combine the items
        elif key in keys_1 and key in keys_2:
            key_2 = case_keys_2[keys_2.index(key)]
            if isinstance(d1[case_key], dict) and isinstance(d2[key_2], dict):
                out[key] = combine_dicts(
                    d1[case_key],
                    d2[key_2],
                    case_sensitive=case_sensitive,
                )
            elif isinstance(d1[case_key], list) and isinstance(d2[key_2], list):
                out[key] = [item for item in [*d1[case_key], *d2[key_2]]]
            else:
                try:
                    out[key] = deepcopy(d1[case_key]) + deepcopy(d2[key_2])
                except TypeError:
                    out[key] = resolve_concatenation(d1[case_key], d2[key_2])

    # go through keys from d2
    if not dict2_conditional:
        for case_key, key in zip(case_keys_2, keys_2):
            # if it's only in d2, add item to out
            if key in keys_2 and key not in keys_1:
                out[case_key] = d2[case_key]

    return out


def subtract_dicts(*args, case_sensitive=False):
    """removes items in d2 from d1 to return a new dictionary
    """
    from copy import deepcopy

    d1 = args[0]
    d2 = args[1:]
    if len(d2) > 1:
        d2 = combine_dicts(
            d2[0],
            *d2[1:],
            case_sensitive=case_sensitive,
            dict2_conditional=False
        )
    else:
        d2 = d2[0]

    out = OrderedDict()
    case_keys_1 = list(d1.keys())
    case_keys_2 = list(d2.keys())
    if case_sensitive:
        keys_1 = case_keys_1
        keys_2 = case_keys_2
    else:
        keys_1 = [
            key.lower() if isinstance(key, str) else key for key in case_keys_1
        ]
        keys_2 = [
            key.lower() if isinstance(key, str) else key for key in case_keys_2
        ]

    # go through keys from d1
    for case_key, key in zip(case_keys_1, keys_1):
        # if the key is only in d1, add the item to out
        if key in keys_1 and key not in keys_2:
            out[case_key] = deepcopy(d1[case_key])
        # if the key is in both, combine the items
        elif key in keys_1 and key in keys_2:
            key_2 = case_keys_2[keys_2.index(key)]
            if isinstance(d1[case_key], dict) and isinstance(d2[key_2], dict):
                out[key] = subtract_dicts(
                    d1[case_key],
                    d2[key_2],
                    case_sensitive=case_sensitive,
                )
            else:
                try:
                    out[key] = deepcopy(d1[case_key]) - deepcopy(d2[key_2])
                except TypeError:
                    out.setdefault(key, type(d1[case_key])())
                    if isinstance(d1[case_key], list):
                        if case_sensitive:
                            for item in d1[case_key]:
                                if d1[case_key] not in d2[key_2]:
                                    out[key].append(d1[case_key])
                        else:
                            items = d2[key_2]
                            if isinstance(items, str) or not hasattr(items, "__iter__"):
                                items = [items]
                            case_items = [x.lower() if isinstance(x, str) else x for x in items]
                            for item in d1[case_key]:
                                if isinstance(item, str) and item.lower() in case_items:
                                    continue
                                elif item in case_items:
                                    continue
                                out[key].append(item)
                    
                    elif hasattr(d1[case_key], "__sub__"):
                        try:
                            out[key] = d1[case_key] - d2[key_2]
                        except TypeError:
                            raise TypeError("cannot subtract %s from %s" % (
                                repr(d2[key_2]), repr(d1[case_key])
                            ))

    return out


def integrate(fun, start, stop, num=101):
    """numerical integration using Simpson's method
    fun - function to integrate
    start - starting point for integration
    stop - stopping point for integration
    num - number of points used for integration"""

    delta_x = float(stop - start) / (num - 1)
    x_set = np.linspace(start, stop, num=num)
    running_sum = -(fun(start) + fun(stop))
    i = 1
    max_4th_deriv = 0
    while i < num:
        if i % 2 == 0:
            running_sum += 2 * fun(x_set[i])
        else:
            running_sum += 4 * fun(x_set[i])

        if i < num - 4 and i >= 3:
            sg_4th_deriv = (
                6 * fun(x_set[i - 3])
                + 1 * fun(x_set[i - 2])
                - 7 * fun(x_set[i - 1])
                - 3 * fun(x_set[i])
                - 7 * fun(x_set[i + 1])
                + fun(x_set[i + 2])
                + 6 * fun(x_set[i + 3])
            )
            sg_4th_deriv /= 11 * delta_x ** 4

            if abs(sg_4th_deriv) > max_4th_deriv:
                max_4th_deriv = abs(sg_4th_deriv)

        i += 1

    running_sum *= delta_x / 3.0

    # close enough error estimate
    e = (abs(stop - start) ** 5) * max_4th_deriv / (180 * num ** 4)

    return (running_sum, e)


def same_cycle(graph, a, b):
    """
    Determines if Atom :a: and Atom :b: are in the same cycle in a undirected :graph:

    Returns: True if cycle found containing a and b, False otherwise

    :graph: connectivity matrix or Geometry
    :a:, :b: indices in connectivity matrix/Geometry or Atoms in Geometry
    """
    from AaronTools.geometry import Geometry

    if isinstance(a, Atoms.Atom):
        a = graph.atoms.index(a)
    if isinstance(b, Atoms.Atom):
        b = graph.atoms.index(b)
    if isinstance(graph, Geometry):
        graph = [
            [graph.atoms.index(j) for j in i.connected] for i in graph.atoms
        ]
    graph = [[i for i in j] for j in graph]

    graph, removed = trim_leaves(graph)

    if a in removed or b in removed:
        return False
    path = shortest_path(graph, a, b)
    for p, q in zip(path[:-1], path[1:]):
        graph[p].remove(q)
        graph[q].remove(p)

    path = shortest_path(graph, a, b)
    if path is None:
        return False
    return True


def shortest_path(graph, start, end):
    """
    Find shortest path from :start: to :end: in :graph: using Dijkstra's algorithm
    Returns: list(node_index) if path found, None if path not found

    :graph: the connection matrix or Geometry
    :start: the first atom or node index
    :end: the last atom or node index
    """
    from AaronTools.geometry import Geometry

    if isinstance(start, Atoms.Atom):
        start = graph.atoms.index(start)
    if isinstance(end, Atoms.Atom):
        end = graph.atoms.index(end)
    if isinstance(graph, Geometry):
        graph = [
            [graph.atoms.index(j) for j in i.connected if j in graph.atoms]
            for i in graph.atoms
        ]
    graph = [[i for i in j] for j in graph]

    # initialize distance array, parent array, and set of unvisited nodes
    dist = [np.inf for x in graph]
    parent = [-1 for x in graph]
    unvisited = set([i for i in range(len(graph))])

    dist[start] = 0
    current = start
    while True:
        # for all unvisited neighbors of current node, update distances
        # if we update the distance to a neighboring node,
        # then also update its parent to be the current node
        for v in graph[current]:
            if v not in unvisited:
                continue
            if dist[v] == np.inf:
                new_dist = dist[current] + 1
            else:
                new_dist = dist[current] + dist[v]
            if dist[v] > new_dist:
                dist[v] = new_dist
                parent[v] = current
        # mark current node as visited
        # select closest unvisited node to be next node
        # break loop if we found end node or if no unvisited connected nodes
        unvisited.remove(current)
        if end not in unvisited:
            break
        current = None
        for u in unvisited:
            if current is None or dist[u] < dist[current]:
                current = u
        if dist[current] == np.inf:
            break
    # return shortest path from start to end
    path = [end]
    while True:
        if parent[path[-1]] == -1:
            break
        path += [parent[path[-1]]]
    path.reverse()
    if path[0] != start or path[-1] != end:
        return None
    return path


def trim_leaves(graph, _removed=None):
    from AaronTools.geometry import Geometry

    # print(_removed)
    if _removed is None:
        _removed = []

    if isinstance(graph, Geometry):
        graph = [
            [graph.atoms.index(j) for j in i.connected] for i in graph.atoms
        ]
    graph = [[i for i in j] for j in graph]
    some_removed = False

    for i, con in enumerate(graph):
        if len(con) == 1:
            graph[con[0]].remove(i)
            graph[i].remove(con[0])
            some_removed = True
            _removed += [i]

    if some_removed:
        graph, _removed = trim_leaves(graph, _removed)

    return graph, set(_removed)


def to_closing(string, brace):
    """returns the portion of string from the beginning to the closing
    paratheses or bracket denoted by brace
    brace can be '(', '{', or '['
    if the closing paratheses is not found, returns None instead"""
    if brace == "(":
        pair = ("(", ")")
    elif brace == "{":
        pair = ("{", "}")
    elif brace == "[":
        pair = ("[", "]")
    else:
        raise RuntimeError("brace must be '(', '{', or '['")

    out = ""
    count = 0
    for x in string:
        if x == pair[0]:
            count += 1
        elif x == pair[1]:
            count -= 1

        out += x
        if count == 0:
            break

    if count != 0:
        return None
    else:
        return out


def rotation_matrix(theta, axis, renormalize=True):
    """rotation matrix for rotating theta radians about axis"""
    # I've only tested this for rotations in R3
    dim = len(axis)
    if renormalize:
        if np.linalg.norm(axis) == 0:
            axis = np.zeros(dim)
            axis[0] = 1.0
        axis = axis / np.linalg.norm(axis)
    outer_prod = np.outer(axis, axis)
    cos_comp = cos(theta)
    outer_prod *= 1 - cos_comp
    iden = np.identity(dim)
    cos_comp = iden * cos_comp
    sin_comp = sin(theta) * (np.ones((dim, dim)) - iden)
    cross_mat = np.zeros((dim, dim))
    for i in range(0, dim):
        for j in range(0, i):
            p = 1
            if (i + j) % 2 != 0:
                p = -1

            cross_mat[i][j] = -1 * (p * axis[dim - (i + j)])
            cross_mat[j][i] = p * axis[dim - (i + j)]

    return outer_prod + cos_comp + sin_comp * cross_mat


def mirror_matrix(norm, renormalize=True):
    """mirror matrix for the specified norm"""
    if renormalize:
        norm /= np.linalg.norm(norm)
    A = np.identity(3)
    B = -2 * np.outer(norm, norm)
    return A + B


def fibonacci_sphere(radius=1, center=np.zeros(3), num=500):
    """
    returns a grid of points that are equally spaced on a sphere
    with the specified radius and center
    number of points can be adjusted with num
    """
    # generate a grid of points on the unit sphere
    grid = np.zeros((num, 3))
    
    i = np.arange(0, num) + 0.5
    phi = np.arccos(1 - 2 * i / num)
    ratio = (1 + sqrt(5.)) / 2
    theta = 2 * np.pi * i / ratio
    grid[:, 0] = np.cos(theta) * np.sin(phi)
    grid[:, 1] = np.sin(theta) * np.sin(phi)
    grid[:, 2] = np.cos(phi)

    # scale the points to the specified radius and move the center
    grid *= radius
    grid += center

    return grid


def lebedev_sphere(radius=1, center=np.zeros(3), num=302):
    """
    returns one of the Lebedev grid points (xi, yi, zi)
    and weights (wi) with the specified radius and center.
    Weights do not include r**2, so integral of F(x,y,z)
    over sphere is 4*pi*r**2\sum_i{F(xi,yi,zi)wi}.  The number
    of points (num) must be one of 110, 194, 302, 590, 974, 1454, 2030, 2702, 5810
    """
    try:
        grid, weights = _loaded_leb_grids[num]
        return radius * grid + center, weights.copy()
    except KeyError:
        pass
    
    # read grid data  on unit sphere
    grid_file = os.path.join(
        AARONTOOLS, "utils", "quad_grids", "Leb" + str(num) + ".grid"
    )
    if not os.path.exists(grid_file):
        # maybe some other error type?
        raise NotImplementedError(
            "cannot use Lebedev grid with %i points\n" % num
            + "use one of 110, 194, 302, 590, 974, 1454, 2030, 2702, 5810"
        )
    grid_data = np.loadtxt(grid_file)
    grid = grid_data[:, :3]
    weights = grid_data[:, 3]
    _loaded_leb_grids[num] = (grid, weights.copy())

    # scale the points to the specified radius and move the center
    grid = grid * radius + center

    return grid, weights


def gauss_legendre_grid(start=-1, stop=1, num=32):
    """
    returns a Gauss-Legendre grid points (xi) and weights
    (wi)for the range start to stop. Integral over F(x) is
    \sum_i{F(xi)wi}. The number of points (num) must be one
    of 20, 32, 64, 75, 99, 127
    """
    # read grid points on the range [-1,1] and weights
    grid_file = os.path.join(
        AARONTOOLS, "utils", "quad_grids", "Leg" + str(num) + ".grid"
    )
    if not os.path.exists(grid_file):
        # maybe some other error type?
        raise NotImplementedError(
            "cannot use Gauss-Legendre grid with %i points\n" % num
            + "use one of 20, 32, 64, 75, 99, 127"
        )
    grid_data = np.loadtxt(grid_file)

    # shift grid range to [start, stop]
    grid = grid_data[:, 0] * (stop - start) / 2 + start + (stop - start) / 2

    # adjust weights for new range
    weights = grid_data[:, 1] * (stop - start) / 2

    return grid, weights


def perp_vector(vec):
    """
    returns a vector orthonormal to vec (np.ndarray)
    if vec is 2D, returns a vector orthonormal to the
    plane of best for the rows of vec
    """
    vec = np.squeeze(vec)
    if vec.ndim == 1:
        out_vec = np.roll(vec, 1)
        if all(x == vec[0] for x in vec):
            for k in range(0, len(vec)):
                if out_vec[k] != 0:
                    out_vec[k] *= -1
                    break
        if np.linalg.norm(vec) == 0:
            # a zero-vector was given
            return np.ones(len(vec)) / len(vec)

        out = np.cross(out_vec, vec)
        out /= np.linalg.norm(out)
        return out

    elif vec.ndim == 2:
        xyz = vec - np.mean(vec, axis=0)
        cov_prod = np.dot(xyz.T, xyz)
        u, s, vh = np.linalg.svd(cov_prod, compute_uv=True)
        return u[:, -1]

    raise NotImplementedError(
        "cannot determine vector perpendicular to %i-dimensional array"
        % vec.ndim
    )


def get_filename(path, include_parent_dir=True):
    """returns the name of the file without parent directories or extension"""
    if not include_parent_dir:
        fname = os.path.basename(path)
    else:
        fname = path
    fname, _ = os.path.splitext(fname)
    return fname


def boltzmann_coefficients(energies, temperature, absolute=True):
    """
    returns boltzmann weights for the energies and T
    energies - numpy array of energies in kcal/mol
    temperature - T in K
    absolute - True if the energies given are absolute, false if they are relative energies
    """
    if absolute:
        min_nrg = min(energies)
        energies -= min_nrg
    weights = np.exp(-energies / (PHYSICAL.R * temperature))
    return weights


def boltzmann_average(energies, values, temperature, absolute=True):
    """
    returns the AVT result for the values corresponding to the energies
    energies - np.array, energy for each state in kcal/mol
    values - np.array, values for which the weighting is applied; the ith value corresponds to the ith energy
    temperature - float, temperature in K
    absolute - True if the energies given are absolute, false if they are relative energies
    """
    weights = boltzmann_coefficients(energies, temperature, absolute=absolute)
    avg = np.dot(weights, values) / sum(weights)
    return avg


def glob_files(infiles, parser=None):
    """
    globs input files
    used for command line scripts because Windows doesn't support globbing...
    """
    from glob import glob
    import sys

    if isinstance(infiles, str):
        infiles = [infiles]

    outfiles = []
    for f in infiles:
        if isinstance(f, str):
            outfiles.extend(glob(f))
        elif len(sys.argv) > 1:
            outfiles.append(f)

    if not outfiles and all(isinstance(f, str) for f in infiles):
        raise RuntimeError(
            "no files could be found for %s" % ", ".join(infiles)
        )
    elif not outfiles and parser is not None:
        parser.print_help()
        sys.exit(0)

    return outfiles


def angle_between_vectors(v1, v2, renormalize=True):
    """returns the angle between v1 and v2 (numpy arrays)"""
    if renormalize:
        v1 = v1 / np.linalg.norm(v1)
        v2 = v2 / np.linalg.norm(v2)

    v12 = v2 - v1
    c2 = np.dot(v12, v12)
    t = (c2 - 2) / -2
    if t > 1:
        t = 1
    elif t < -1:
        t = -1

    # math.acos is faster than numpy.arccos for non-arrays
    return acos(t)


def is_alpha(test):
    rv = re.search("^[a-zA-Z]+$", test)
    return bool(rv)


def is_int(test):
    rv = re.search("^[+-]?\d+$", test)
    return bool(rv)


def is_num(test):
    rv = re.search("^[+-]?\d+\.?\d*", test)
    return bool(rv)

def getuser():
    """returns the username of the user"""
    for envar in ["LOGNAME", "USER", "LNAME", "USERNAME"]:
        user = os.getenv(envar, default=False)
        if user:
            return user
    
    try:
        import pwd
        return pwd.getpwuid(os.getuid())[0]
    except (ModuleNotFoundError, ImportError):
        from psutil import username
        return username()

def available_memory():
    """returns available memory in B"""
    for envar in ["SLURM_MEM_PER_NODE"]:
        mem = os.getenv(envar, default=False)
        if mem:
            return 1e6 * int(mem)
    
    from psutil import virtual_memory
    return virtual_memory().free

float_num = re.compile("[-+]?\d+\.?\d*")



#!/usr/bin/env python3

import sys
from os.path import splitext
import argparse

from AaronTools.geometry import Geometry
from AaronTools.fileIO import FileReader, read_types
from AaronTools.theory import *
from AaronTools.utils.utils import combine_dicts, get_filename, glob_files

theory_parser = argparse.ArgumentParser(
    description="print CREST input files",
    formatter_class=argparse.RawTextHelpFormatter
)

theory_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file",
)

theory_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    required=False,
    dest="outfile",
    help="output destination\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: stdout"
)

theory_parser.add_argument(
    "-if", "--input-format",
    type=str,
    default=None,
    dest="input_format",
    choices=read_types,
    help="file format of input - xyz is assumed if input is stdin",
)

theory_parser.add_argument(
    "-q", "--charge",
    type=int,
    dest="charge",
    default=None,
    help="net charge\nDefault: 0 or what is found in the input file",
)

theory_parser.add_argument(
    "-mult", "--multiplicity",
    type=int,
    dest="multiplicity",
    default=None,
    help="electronic multiplicity\nDefault: 1 or what is found in the input file",
)

theory_parser.add_argument(
    "-p", "--cores",
    type=int,
    dest="processors",
    default=None,
    required=False,
    help="number of cpu cores to use",
)

theory_options = theory_parser.add_argument_group("Theory options")
theory_options.add_argument(
    "-m", "--method",
    type=str,
    dest="method",
    required=False,
    default="GFN2-xTB",
    help="method (e.g. GFN-FF)\nDefault: GFN2-xTB",
)

theory_options.add_argument(
    "-sv", "--solvent",
    required=False,
    default=None,
    dest="solvent",
    help="solvent",
)

theory_options.add_argument(
    "-sm", "--solvent-model",
    required=False,
    default=None,
    dest="solvent_model",
    help="implicit solvent model",
)


search_options = theory_parser.add_argument_group("Search options")
search_options.add_argument(
    "-ca", "--constrained-atoms",
    nargs=1,
    type=str,
    action="append",
    default=None,
    dest="atoms",
    help="comma- or hyphen-separated list of atoms (1-indexed) to constrain during optimization",
)

search_options.add_argument(
    "-cb", "--constrain-bond",
    nargs=1,
    action="append",
    default=None,
    dest="bonds",
    help="list of comma-separated atom pairs\n" +
    "the distance between the atoms in each pair will be constrained during optimization",
)

search_options.add_argument(
    "-cang", "--constrain-angle",
    type=str,
    nargs=1,
    action="append",
    default=None,
    dest="angles",
    help="list of comma-separated atom trios\n" +
    "the angle defined by each trio will be constrained during optimization",
)

search_options.add_argument(
    "-ct", "--constrain-torsion",
    type=str,
    nargs=1,
    action="append",
    default=None,
    dest="torsions",
    help="list of comma-separated atom quartets\n" +
    "the torsional angle defined by each quartet will be constrained during optimization",
)


gaussian_options = theory_parser.add_argument_group("CREST options")
gaussian_options.add_argument(
    "-xc", "--xcontrol",
    action="append",
    nargs="+",
    default=[],
    dest=XTB_CONTROL_BLOCKS,
    metavar=("KEYWORD", "OPTION"),
    help="xcontrol options\nexample: --metadyn coord original.xyz\n" +
    "input file(s) should not be right after --xcontrol",
)

gaussian_options.add_argument(
    "-cmd", "--command",
    action="append",
    nargs="+",
    default=[],
    dest=XTB_COMMAND_LINE,
    metavar=("COMMAND", "VALUE"),
    help="command line options (without --)\n" +
    "example: --command tautomerize\ninput file(s) should not be right after --link0",
)


args = theory_parser.parse_args()

kwargs = {}

for pos in [
        XTB_COMMAND_LINE, XTB_CONTROL_BLOCKS
    ]:
    opts = getattr(args, pos)
    if opts:
        if pos not in kwargs:
            kwargs[pos] = {}

        for opt in opts:
            setting = opt.pop(0)
            if setting not in kwargs[pos]:
                kwargs[pos][setting] = []

            kwargs[pos][setting].extend(opt)

# Theory() is made for each file because we might be using things from the input file
for f in glob_files(args.infile, parser=theory_parser):
    if isinstance(f, str):
        if args.input_format is not None:
            infile = FileReader((f, args.input_format[0], None), just_geom=False, get_all=True)
        else:
            infile = FileReader(f, just_geom=False, get_all=True)
    else:
        if args.input_format is not None:
            infile = FileReader(
                ("from stdin", args.input_format[0], f), just_geom=False, get_all=True
            )
        else:
            if len(sys.argv) >= 1:
                infile = FileReader(("from stdin", "xyz", f), just_geom=False, get_all=True)

    geom = Geometry(infile, refresh_ranks=False)


    if args.solvent is not None or args.solvent_model is not None:
        if args.solvent_model is None or args.solvent is None:
            raise RuntimeError("--solvent and --solvent-model must both be specified")

        solvent = ImplicitSolvent(args.solvent_model, args.solvent)

    else:
        solvent = None


        constraints = {}
        if args.atoms is not None:
            constraints["atoms"] = []
            for constraint in args.atoms:
                constraints["atoms"].extend(geom.find(constraint))

        if args.bonds is not None:
            constraints["bonds"] = []
            for bond in args.bonds:
                bonded_atoms = geom.find(bond)
                if len(bonded_atoms) != 2:
                    raise RuntimeError(
                        "not exactly 2 atoms specified in a bond constraint\n" +
                        "use the format --constrain-bond 1,2"
                    )
                constraints["bonds"].append(bonded_atoms)

        if args.angles is not None:
            constraints["angles"] = []
            for angle in args.angle:
                angle_atoms = geom.find(angle)
                if len(angle_atoms) != 3:
                    raise RuntimeError(
                        "not exactly 3 atoms specified in a angle constraint\n" +
                        "use the format --constrain-angle 1,2,3"
                    )
                constraints["angles"].append(angle_atoms)

        if args.torsions is not None:
            constraints["torsions"] = []
            for torsion in args.torsions:
                torsion_atoms = geom.find(torsion)
                if len(torsion_atoms) != 4:
                    raise RuntimeError(
                        "not exactly 4 atoms specified in a torsion constraint\n" +
                        "use the format --constrain-torsion 1,2,3,4"
                    )
                constraints["torsions"].append(torsion_atoms)

        if not constraints.keys():
            constraints = None

        job_type = ConformerSearchJob(constraints=constraints)

    if args.charge is None:
        if "charge" in infile.other:
            charge = infile.other["charge"]
        else:
            charge = 0
    else:
        charge = args.charge


    if args.multiplicity is None:
        if "multiplicity" in infile.other:
            multiplicity = infile.other["multiplicity"]
        else:
            multiplicity = 1
    else:
        multiplicity = args.multiplicity

    other_kwargs = {}

    theory = Theory(
        geometry=geom,
        method=args.method,
        solvent=solvent,
        job_type=job_type,
        charge=charge,
        multiplicity=multiplicity,
        processors=args.processors,
        **kwargs,
    )

    if args.outfile:
        outfile = args.outfile
        if "$INFILE" in outfile:
            outfile = outfile.replace("$INFILE", get_filename(f))
        warnings = geom.write(
            append=True,
            outfile=outfile,
            style="crest",
            theory=theory,
            return_warnings=True,
            **other_kwargs
        )
    else:
        out, warnings = geom.write(
            append=True,
            outfile=False,
            style="crest",
            theory=theory,
            return_warnings=True,
            **other_kwargs
        )
        for key, item in out.items():
            print("<<--%s-->>" % key)
            print(item)
    
    if warnings:
        for warning in warnings:
            Geometry.LOG.warning(warning)



import numpy as np

import scipy.interpolate
import scipy.integrate


class Pathway:
    """
    Cartesian interpolation of a molecule's coordinates
    """
    def __init__(
        self,
        coordinates,
        interpolation="clampedcubic",
        other_vars=None,
        x_vals=None,
        weights=None,
        **interpolator_kwargs
    ):
        self.coords = coordinates
        self.interpolation = interpolation
        self.other_vars = other_vars
        self._interpolation = None
        self._other_vars_interpolation = None
        self._region_lengths = None
        self.weights = weights
        
        if x_vals is None:
            x_vals = np.arange(0, len(self.coords))
        
        self.x_vals = x_vals
        
        self._build_interpolation(**interpolator_kwargs)
    
    def _build_interpolation(self, **interpolator_kwargs):
        n_nodes = len(self.coords)
        n_atoms = len(self.coords[0])
        coords = np.reshape(self.coords, (n_nodes, 3 * n_atoms))
        if self.weights is None:
            self.weights = np.ones(n_atoms)
        weights = np.repeat(self.weights, 3)
        for i in range(0, n_nodes):
            coords[i] *= weights

        if self.other_vars:
            self._other_vars_interpolation = dict()

        if self.interpolation.lower().startswith("akima"):
            interp_func = scipy.interpolate.Akima1DInterpolator
        
        elif self.interpolation.lower().startswith("pchip"):
            interp_func = scipy.interpolate.PchipInterpolator
        
        elif self.interpolation.lower().startswith("naturalcubic"):
            interp_func = lambda *args: scipy.interpolate.CubicSpline(
                *args, bc_type="natural"
            )

        elif self.interpolation.lower().startswith("clampedcubic"):
            interp_func = lambda *args: scipy.interpolate.CubicSpline(
                *args, bc_type="clamped"
            )
        
        elif self.interpolation.lower().startswith("nakcubic"):
            interp_func = lambda *args: scipy.interpolate.CubicSpline(
                *args, bc_type="not-a-knot"
            )
        
        else:
            raise RuntimeError(
                "interpolation method not known: %s" % self.interpolation
            )
        
        self._interpolation = [
            interp_func(self.x_vals, coord, **interpolator_kwargs) for coord in coords.T
        ]
        if self.other_vars:
            for var, vals in self.other_vars.items():
                self._other_vars_interpolation[var] = interp_func(
                    self.x_vals,
                    vals,
                )
    
        deriv = [interp.derivative(1) for interp in self._interpolation]
        region_starts = self.x_vals[:-1]
        region_stops = self.x_vals[1:]
        self._region_lengths = np.zeros(n_nodes - 1)
        n_pts = 33
        for i in range(0, len(region_starts)):
            total = 0
            a = region_starts[i]
            b = region_stops[i]
            for k, x in enumerate(
                np.linspace(a, b, num=n_pts)
            ):
                region_length = 0
                for j in range(0, 3 * n_atoms):
                    val = deriv[j](x) ** 2
                    if k == 0 or k == n_pts - 1:
                        region_length += val
                    elif k % 2 == 0:
                        region_length += 2 * val
                    else:
                        region_length += 4 * val
                
                total += np.sqrt(region_length)
            
            total = (b - a) * total / (n_pts * 6)
            self._region_lengths[i] = total

    def s_at_t(self, t):
        total_length = sum(self._region_lengths)
        u = t * total_length
        r = 0
        for i in range(0, len(self._region_lengths)):
            if u > sum(self._region_lengths[:i]):
                r = i
    
        # print("point is in region %i" % r)
        s = self.x_vals[r]
        s += (self.x_vals[r + 1] - self.x_vals[r]) * (u - sum(self._region_lengths[:r])) / self._region_lengths[r]
        return min(s, max(self.x_vals))

    def interpolate_coords(self, t):
        """
        returns array for the interpolated pathway
        at point t
        t should be between 0 and 1
        """
        s = self.s_at_t(t)
        # print("getting coordinates at s=%.4f (t=%.2f)" % (s, t))
        interpolated_coords = np.array([
            coord(s) for coord in self._interpolation
        ])
        interpolated_coords
        interpolated_coords = interpolated_coords.reshape(
            (len(interpolated_coords) // 3, 3)
        )
        return interpolated_coords / self.weights[:, np.newaxis]
    
    def interpolate_geometry(self, t, geom):
        geom = geom.copy()
        coords = self.interpolate_coords(t)
        geom.coords = coords
        return geom
    
    def interpolate_other_var(self, t, var):
        s = self.s_at_t(t)
        return self._other_vars_interpolation[var](s)

    def other_var_derivative(self, t, var):
        s = self.s_at_t(t)
        return self._other_vars_interpolation[var].derivative(1)(s)
    
    def tangent(self, t):
        s = self.s_at_t(t)
        deriv = [interp.derivative(1) for interp in self._interpolation]
        tangent = np.array([f(s) for f in deriv])
        tangent = tangent.reshape(
            (len(tangent) // 3, 3)
        )
        return tangent / self.weights[:, np.newaxis]



import configparser
from difflib import SequenceMatcher as seqmatch
import itertools as it
import os
import re

import AaronTools
from AaronTools import addlogger
from AaronTools.const import AARONLIB, AARONTOOLS
from AaronTools.theory import (
    GAUSSIAN_COMMENT,
    GAUSSIAN_CONSTRAINTS,
    GAUSSIAN_COORDINATES,
    GAUSSIAN_GEN_BASIS,
    GAUSSIAN_GEN_ECP,
    GAUSSIAN_POST,
    GAUSSIAN_PRE_ROUTE,
    GAUSSIAN_ROUTE,
    ORCA_BLOCKS,
    ORCA_COMMENT,
    ORCA_COORDINATES,
    ORCA_ROUTE,
    PSI4_AFTER_JOB,
    PSI4_BEFORE_GEOM,
    PSI4_BEFORE_JOB,
    PSI4_COMMENT,
    PSI4_JOB,
    PSI4_MOLECULE,
    PSI4_OPTKING,
    PSI4_SETTINGS,
    PSI4_SOLVENT,
    SQM_COMMENT,
    SQM_QMMM,
    QCHEM_MOLECULE,
    QCHEM_REM,
    QCHEM_COMMENT,
    QCHEM_SETTINGS,
    XTB_CONTROL_BLOCKS,
    XTB_COMMAND_LINE,
    Theory,
)
from AaronTools.theory.implicit_solvent import ImplicitSolvent
from AaronTools.theory.job_types import job_from_string
from AaronTools.utils.utils import getuser


THEORY_OPTIONS = [
    "GAUSSIAN_COMMENT",
    "GAUSSIAN_CONSTRAINTS",
    "GAUSSIAN_COORDINATES",
    "GAUSSIAN_GEN_BASIS",
    "GAUSSIAN_GEN_ECP",
    "GAUSSIAN_POST",
    "GAUSSIAN_PRE_ROUTE",
    "GAUSSIAN_ROUTE",
    "ORCA_BLOCKS",
    "ORCA_COMMENT",
    "ORCA_COORDINATES",
    "ORCA_ROUTE",
    "PSI4_AFTER_JOB",
    "PSI4_BEFORE_GEOM",
    "PSI4_BEFORE_JOB",
    "PSI4_COMMENT",
    "PSI4_MOLECULE",
    "PSI4_JOB",
    "PSI4_OPTKING",
    "PSI4_SETTINGS",
    "SQM_COMMENT",
    "SQM_QMMM",
    "QCHEM_MOLECULE",
    "QCHEM_REM",
    "QCHEM_COMMENT",
    "QCHEM_SETTINGS",
    "XTB_COMMAND_LINE",
    "XTB_CONTROL_BLOCKS"
]


@addlogger
class Config(configparser.ConfigParser):
    """
    Reads configuration information from INI files found at:
        $QCHASM/AaronTools/config.ini
        $AARONLIB/config.ini
        ./config.ini or /path/to/file supplied during initialization
    Access to configuration information available using dictionary notation.
        eg: self[`section_name`][`option_name`] returns `option_value`
    See help(configparser.ConfigParser) for more information
    """

    LOG = None
    SPEC_ATTRS = [
        "_changes",
        "_changed_list",
        "_args",
        "_kwargs",
        "infile",
        "metadata",
    ]

    @classmethod
    def _process_content(cls, filename, quiet=True):
        """
        process file content to handle optional default section header
        """
        contents = filename
        if os.path.isfile(filename):
            try:
                with open(filename) as f:
                    contents = f.read()
            except Exception as e:
                if not quiet:
                    cls.LOG.INFO("failed to read %s: %s", filename, e)
                return ""
        elif not len(filename.splitlines()) > 1:
            return ""
        try:
            configparser.ConfigParser().read_string(contents)
        except configparser.MissingSectionHeaderError:
            contents = "[DEFAULT]\n" + contents
        return contents

    def __init__(
        self, 
        infile=None, 
        quiet=False, 
        skip_user_default=False, 
        interpolation=configparser.ExtendedInterpolation(), 
        **kwargs
    ):
        """
        infile: the configuration file to read
        quiet: prints helpful status information
        skip_user_default: change to True to skip importing user's default config files
        **kwargs: passed to initialization of parent class
        """
        configparser.ConfigParser.__init__(
            self, interpolation=None, comment_prefixes=("#"), **kwargs
        )
        self.infile = infile
        if not quiet:
            print("Reading configuration...")
        self._read_config(infile, quiet, skip_user_default)

        # enforce case-sensitivity in certain sections
        for section in self:
            if section in [
                "Substitution",
                "Mapping",
                "Configs",
                "Results",
            ]:
                continue
            for option, value in list(self[section].items()):
                if section == "Geometry" and option.lower().startswith(
                    "structure"
                ):
                    del self[section][option]
                    option = option.split(".")
                    option[0] = option[0].lower()
                    option = ".".join(option)
                    self[section][option] = value
                    continue
                if section == "Geometry" and "structure" in value.lower():
                    re.sub("structure", "structure", value, flags=re.I)
                    self[section][option] = value
                if option.lower() != option:
                    self[section][option.lower()] = value
                    del self[section][option]

        # handle included sections
        self._parse_includes()
        if infile is not None:
            self.read(infile)

        # set additional default values
        if infile:
            if "top_dir" not in self["DEFAULT"]:
                self["DEFAULT"]["top_dir"] = os.path.dirname(
                    os.path.abspath(infile)
                )
            if "name" not in self["DEFAULT"]:
                self["DEFAULT"]["name"] = ".".join(
                    os.path.relpath(
                        infile, start=self["DEFAULT"]["top_dir"]
                    ).split(".")[:-1]
                )
        else:
            if "top_dir" not in self["DEFAULT"]:
                self["DEFAULT"]["top_dir"] = os.path.abspath(os.path.curdir)

        # handle substitutions/mapping
        self._changes = {}
        self._changed_list = []
        self._parse_changes()

        # for passing to Theory(*args, **kwargs)
        self._args = []
        self._kwargs = {}

        # metadata is username and project name
        self.metadata = {
            "user": self.get(
                "DEFAULT",
                "user",
                fallback=self.get("HPC", "user", fallback=getuser()),
            ),
            "project": self.get("DEFAULT", "project", fallback=""),
        }
    
    def get(self, section, option, *, junk="_ ", max_junk=1, **kwargs):
        """
        see ConfigParser.get for details
        junk are characters that are not important in the option name
        max_junk - number of allowed junk characters
        e.g. junk="_" with max_junk=1 when looking for
        'empirical dispersion' will match 'empirical_dispersion'
        """
        # dummy class to allow user to specify whatever they
        # want for fallback - even False and None
        class NoFallback:
            pass

        fallback = NoFallback
        if "fallback" in kwargs:
            fallback = kwargs.pop("fallback")
        
        # see if option is present as given
        out = super().get(section, option, fallback=NoFallback, **kwargs)
        # otherwise, look through the rest of the options to see if one
        # is basically the same but with some extra junk characters
        # e.g. 'empirical_dispersion' instead of 'empirical dispersion'
        if out is NoFallback and junk and self.has_section(section):
            for test_opt in self.options(section):
                # calculate similarity
                similarity = seqmatch(lambda x: x in junk, option, test_opt).ratio()
                # must have no more than max_junk junk characters
                if similarity >= 1 - max_junk / len(option) and any(
                    x in test_opt for x in junk
                ):
                    out = super().get(section, test_opt, fallback=NoFallback, **kwargs)
                    break
        
        if out is NoFallback and fallback is NoFallback:
            raise configparser.NoOptionError(section, option)
        if out is NoFallback:
            out = fallback
        return out

    def optionxform(self, option):
        return str(option)

    def __str__(self):
        rv = ""
        for section in self:
            if "." in section:
                continue
            rv += "[{}]\n".format(section)
            for option, value in self[section].items():
                rv += "{} = {}\n".format(option, value)
            rv += "\n"
        return rv

    def copy(self):
        config = Config(infile=None, quiet=True)
        for section in config.sections():
            config.remove_section(section)
        for option in list(config["DEFAULT"].keys()):
            config.remove_option("DEFAULT", option)
        for section in ["DEFAULT"] + self.sections():
            try:
                config.add_section(section)
            except (configparser.DuplicateSectionError, ValueError):
                pass
            for key, val in self[section].items():
                config[section][key] = val
        for section in self.SPEC_ATTRS:
            setattr(config, section, getattr(self, section))
        return config

    def for_change(self, change, structure=None):
        this_config = self.copy()
        if structure is not None:
            this_config["Job"]["name"] = structure.name
        if change:
            this_config["Job"]["name"] = os.path.join(
                change, this_config["Job"]["name"]
            )
        this_config._changes = {change: self._changes[change]}
        return this_config

    def _parse_changes(self):
        for section in ["Substitution", "Mapping"]:
            if section not in self:
                continue
            if self[section].getboolean("reopt", fallback=True):
                self._changes[""] = ({}, None)
            for key, val in self[section].items():
                if key in self["DEFAULT"]:
                    continue
                del self[section][key]
                key = "\n".join(["".join(k.split()) for k in key.split("\n")])
                val = "\n".join(["".join(v.split()) for v in val.split("\n")])
                self[section][key] = val
            for key, val in self[section].items():
                if key in self["DEFAULT"] or key == "reopt":
                    continue
                if "=" not in val:
                    val = [v.strip() for v in val.split(",")]
                else:
                    tmp = [v.strip() for v in val.split(";")]
                    val = []
                    for t in tmp:
                        t = t.strip()
                        if not t:
                            continue
                        elif "\n" in t:
                            val += t.split("\n")
                        else:
                            val += [t]
                    tmp = {}
                    for i, v in enumerate(val):
                        if i == 0 and len(v.split("=")) == 1:
                            v = "{}={}".format(key, v)
                            val[i] = v
                            del self[section][key]
                            key = ""
                            self[section]["~PLACEHOLDER~"] = ";".join(val)
                        v = v.split("=")
                        if (
                            not key.startswith("&combinations")
                            and "(" not in v[0]
                        ):
                            v[0] = v[0].split(",")
                        else:
                            v[0] = [v[0]]
                        for k in v[0]:
                            tmp[k] = v[1]
                    val = tmp
                # handle request for combinations
                if key.startswith("&combination"):
                    atoms = []
                    subs = []
                    # val <= { "2, 4": "H, CH3", "7, 9": "OH, NH2", .. }
                    for k, v in val.items():
                        if "(" not in k:
                            # regular substituents
                            atoms.append(k.split(","))
                        else:
                            # ring substitutions
                            atoms.append(re.findall("\(.*?\)", k))
                        subs.append([None] + [i for i in v.strip().split(",")])
                    # atoms <= [ [2, 4],         [7, 9],      .. ]
                    # subs  <= [ [None, H, CH3], [None, OH, NH2], .. ]
                    for combo in it.product(*[range(len(s)) for s in subs]):
                        # combos <= (0, 0,..), (0,.., 1),..(1,.. 0),..(1,.., 1),..
                        if not any(combo):
                            # skip if no substitutions
                            # (already included if reopt=True)
                            continue
                        name = []
                        tmp = {}
                        for i, p in enumerate(combo):
                            # don't add subsitution if sub == None
                            if subs[i][p] is None:
                                continue
                            name.append(subs[i][p])
                            for a in atoms[i]:
                                tmp[a] = subs[i][p]
                        name = "_".join(name)
                        self._changes[name] = (
                            tmp,
                            section,
                        )
                else:
                    if isinstance(val, list):
                        name = "_".join(val)
                        val = {key: ",".join(val)}
                    elif not key:
                        name = "_".join(
                            [
                                "_".join([v] * len(k.split(",")))
                                for k, v in val.items()
                            ]
                        )
                        self[section][name] = self[section]["~PLACEHOLDER~"]
                        del self[section]["~PLACEHOLDER~"]
                    else:
                        name = key
                    self._changes[name] = (val, section)

    def parse_functions(self):
        """
        Evaluates functions supplied in configuration file
        Functions indicated by "%{...}"
        Pulls in values of options indicated by $option_name
        Eg:
            ppn = 4
            memory = %{ $ppn * 2 }GB --> memory = 8GB
        """
        func_patt = re.compile("(%{(.*?)})")
        attr_patt = re.compile("\$([a-zA-Z0-9_:]+)")
        for section in ["DEFAULT"] + self.sections():
            # evaluate functions
            for key, val in self[section].items():
                match_list = func_patt.findall(val)
                while match_list:
                    match = match_list.pop()
                    eval_match = match[1]
                    for attr in attr_patt.findall(match[1]):
                        if ":" in attr:
                            from_section, option = attr.split(":")
                        else:
                            option, from_section = attr, section
                        option = self[from_section][option]
                        eval_match = eval_match.replace("$" + attr, option, 1)
                    try:
                        eval_match = eval(eval_match, {})
                    except TypeError as e:
                        raise TypeError(
                            "{} for\n\t[{}]\n\t{} = {}\nin config file. Could not evaluate {}".format(
                                e.args[0], section, key, val, eval_match
                            )
                        )
                    except (NameError, SyntaxError):
                        if attr_patt.findall(eval_match):
                            eval_match = "%{" + eval_match.strip() + "}"
                        else:
                            eval_match = eval_match.strip()
                    val = val.replace(match[0], str(eval_match))
                    self[section][key] = val
                    match_list = func_patt.findall(val)

    def getlist(self, section, option, *args, delim=",", **kwargs):
        """returns a list of option values by splitting on the delimiter specified by delim"""
        raw = self.get(section, option, *args, **kwargs)
        out = [x.strip() for x in raw.split(delim) if len(x.strip()) > 0]
        return out

    def read(self, filename, quiet=True):
        try:
            self.read_string(self._process_content(filename, quiet=quiet))
        except configparser.ParsingError:
            pass

    def _read_config(self, infile, quiet, skip_user_default):
        """
        Reads configuration information from `infile` after pulling defaults
        """
        filenames = [
            os.path.join(AARONTOOLS, "config.ini"),
        ]
        if not skip_user_default:
            filenames += [os.path.join(AARONLIB, "config.ini")]
        if infile:
            filenames += [infile]
        local_only = False
        job_include = None
        for i, filename in enumerate(filenames):
            if not quiet:
                if os.path.isfile(filename):
                    print("    ✓", end="  ")
                else:
                    print("    ✗", end="  ")
                print(filename)
            content = self._process_content(filename)
            self.read(content, quiet=quiet)
            job_include = self.get("Job", "include", fallback=job_include)
            if filename != infile:
                self.remove_option("Job", "include")
            # local_only can only be overridden at the user level if "False" in the system config file
            if i == 0:
                local_only = self["DEFAULT"].getboolean("local_only")
            elif local_only:
                self["DEFAULT"]["local_only"] = str(local_only)
        if "Job" in self:
            type_spec = [
                re.search("(?<!_)type", option) for option in self["Job"]
            ]
        else:
            type_spec = []
        if not any(type_spec):
            self.set("Job", "include", job_include)

    def get_other_kwargs(self, section="Theory"):
        """
        Returns dict() that can be unpacked and passed to Geometry.write along with a theory
        Example:
        [Theory]
        route = pop NBORead
                opt MaxCycle=1000, NoEigenTest
        end_of_file = $nbo RESONANCE NBOSUM E2PERT=0.0 NLMO BNDIDX $end

        this adds opt(MaxCycle=1000,NoEigenTest) pop=NBORead to the route with any other
        pop or opt options being added by the job type

        'two-layer' options can also be specified as a python dictionary
        the following is equivalent to the above example:
        [Theory]
        route = {"pop":["NBORead"], "opt":["MaxCycle=1000", NoEigenTest"]}
        end_of_file = $nbo RESONANCE NBOSUM E2PERT=0.0 NLMO BNDIDX $end
        """
        # these need to be dicts
        two_layer = [
            GAUSSIAN_ROUTE,
            GAUSSIAN_PRE_ROUTE,
            ORCA_BLOCKS,
            PSI4_JOB,
            QCHEM_REM,
            QCHEM_SETTINGS,
            PSI4_SOLVENT,
            XTB_CONTROL_BLOCKS,
            XTB_COMMAND_LINE,
        ]

        # these need to be dicts, but can only have one value
        two_layer_single_value = [
            PSI4_OPTKING,
            PSI4_SETTINGS,
            PSI4_MOLECULE,
        ]

        # these need to be lists
        one_layer = [
            GAUSSIAN_COMMENT,
            GAUSSIAN_CONSTRAINTS,
            GAUSSIAN_POST,
            ORCA_COMMENT,
            ORCA_ROUTE,
            PSI4_AFTER_JOB,
            PSI4_BEFORE_GEOM,
            PSI4_BEFORE_JOB,
            PSI4_COMMENT,
            QCHEM_MOLECULE,
            QCHEM_COMMENT,
        ]

        theory_kwargs = [
            "method",
            "charge",
            "multiplicity",
            "type",
            "basis",
            "ecp",
            "grid",
            "empirical_dispersion",
        ]

        # two layer options are separated by newline
        # individual options are split on white space, with the first defining the primary layer
        out = {}
        for option in two_layer:
            value = self[section].get(option, fallback=False)
            value = self._kwargs.get(option, value)
            if value:
                if isinstance(value, dict):
                    out[option] = value
                elif "{{" not in value and "{" in value:
                    # if it's got brackets, it's probably a python-looking dictionary
                    # eval it instead of parsing
                    # double brackets would indicate an interpolated value
                    # e.g. link0=chk {{ name }}.chk
                    out[option] = eval(value, {})
                else:
                    out[option] = {}
                    for v in value.splitlines():
                        data = v.split()
                        key = data[0]
                        if len(data) > 1:
                            i = 1
                            info = []
                            while i < len(data):
                                word = data[i]
                                if word == "{{" and i < len(data) - 2:
                                    word += " " + data[i + 1] + " " + data[i + 2]
                                    info.extend(word.split(","))
                                    i += 3
                                    continue
                                info.extend(word.split(","))
                                i += 1
                        else:
                            info = []

                        out[option][key] = [x.strip() for x in info]

        for option in two_layer_single_value:
            value = self.get(section, option, fallback=False)
            value = self._kwargs.get(option, value)
            if value:
                if "{" in value:
                    out[option] = eval(value, {})
                else:
                    out[option] = {}
                    for v in value.splitlines():
                        key = v.split()[0]
                        if len(v.split()) > 1:
                            info = [v.split()[1]]
                        else:
                            info = []

                        out[option][key] = [x.strip() for x in info]

        for option in one_layer:
            value = self[section].get(option, fallback=False)
            value = self._kwargs.get(option, value)
            if value:
                out[option] = value.splitlines()

        for option in theory_kwargs:
            value = self[section].get(option, fallback=False)
            if value:
                out[option] = value

        return out

    def get_constraints(self, geometry):
        constraints = {}
        try:
            con_list = re.findall("\(.*?\)", self["Geometry"]["constraints"])
        except KeyError:
            try:
                geometry.parse_comment()
                con_list = geometry.other["constraint"]
            except KeyError:
                raise RuntimeError(
                    "Constraints for forming/breaking bonds must be specified for TS search"
                )
        for con in con_list:
            tmp = []
            try:
                for c in eval(con, {}):
                    tmp += geometry.find(str(c))
            except TypeError:
                for c in con:
                    tmp += geometry.find(str(c))
            con = [a.name for a in tmp]
            if len(con) == 1:
                constraints.setdefault("atoms", [])
                constraints["atoms"] += [con]
            elif len(con) == 2:
                constraints.setdefault("bonds", [])
                constraints["bonds"] += [con]
            elif len(con) == 3:
                constraints.setdefault("angles", [])
                constraints["angles"] += [con]
            elif len(con) == 4:
                constraints.setdefault("torsions", [])
                constraints["torsions"] += [con]
        return constraints

    def get_theory(self, geometry, section="Theory"):
        """
        Get the theory object according to configuration information
        """
        if not self.has_section(section):
            self.LOG.warning(
                'config has no "%s" section, switching to "Theory"' % section
            )
            section = "Theory"

        kwargs = self.get_other_kwargs(section=section)
        theory = Theory(*self._args, geometry=geometry, **kwargs)
        theory.processors = self["Job"].getint("procs", fallback=None)
        theory.memory = self["Job"].getint("exec_memory", fallback=None)

        # build ImplicitSolvent object
        if self[section].get("solvent", fallback="gas") == "gas":
            theory.solvent = None
        elif self[section]["solvent"]:
            solvent_model = self.get(section, "solvent_model", fallback=False)
            theory.solvent = ImplicitSolvent(
                solvent_model,
                self[section]["solvent"],
            )

        # build JobType list
        job_type = self["Job"].get("type", fallback=False)
        if job_type:
            theory.job_type = []
            numerical = self[section].get("numerical", fallback=False)
            temperature = self[section].get("temperature", fallback=298.15)
            try:
                constraints = self.get_constraints(theory.geometry)
            except RuntimeError:
                constraints = None
            theory.geometry.freeze()
            theory.geometry.relax(self._changed_list)

            info = {
                "numerical": numerical,
                "temperature": temperature,
                "constraints": constraints,
                "geometry": theory.geometry,
            }
            try:
                theory.job_type += [job_from_string(job_type, **info)]
            except ValueError:
                raise ValueError("cannot parse job type: %s" % ".".join(job_type))
        else:
            # default to opt+freq
            theory.job_type = [
                OptimizationJob(geometry=geometry),
                FrequencyJob(
                    numerical=self[section].get("numerical", fallback=False),
                    temperature=self[section].get(
                        "temperature", fallback=None
                    ),
                ),
            ]
        # return updated theory object
        return theory

    def get_template(self):
        # captures name placeholder and iterator from for-loop initilaizer
        for_patt = re.compile("&for\s+(.+)\s+in\s+(.+)")
        # captures structure_dict-style structure/suffix -> (structure['suffix'], suffix)
        parsed_struct_patt = re.compile("(structure\['(\S+?)'\])\.?")
        # captures config-style structure/suffix -> (structure.suffix, suffix)
        structure_patt = re.compile("(structure\.([^\(\s\.]+))\.?")

        def get_multiple(filenames, path=None, suffix=None):
            rv = []
            for name in filenames:
                kind = "Minimum"
                if name.startswith("TS"):
                    kind = "TS"
                if path is not None:
                    name = os.path.join(path, name)
                if not os.path.isfile(name):
                    continue
                geom = AaronTools.geometry.Geometry(name)
                if suffix is not None:
                    geom.name += ".{}".format(suffix)
                rv += [(geom, kind)]
            return rv

        def structure_assignment(line):
            # assignments must be done outside of eval()
            # left -> structure.suffix -> structure_dict["suffix"]
            # right -> eval(right)
            # left = right -> structure_dict[suffix] = eval(right)
            left = line.split("=")[0].strip()
            right = line.split("=")[1].strip()
            suffix_match = parsed_struct_patt.search(left)
            if suffix_match is None:
                raise RuntimeError(
                    "Can only assign to Geometry objects with names of the form `structure.suffix`"
                )
            suffix = suffix_match.group(2)
            structure_dict[suffix] = eval(right, eval_dict)
            structure_dict[suffix].name = ".".join(
                structure_dict[suffix].name.split(".")[:-1] + [suffix]
            )
            if structure_dict[suffix].name.startswith("TS"):
                kind_dict[suffix] = "TS"

        def structure_suffix_parse(line, for_loop=None):
            if for_loop is not None:
                for_match, it_val = for_loop
            for structure_match in structure_patt.findall(line):
                if getattr(AaronTools.geometry.Geometry, structure_match[1], False):
                    continue
                # if our suffix is not the iterator, keep it's value for the dict key
                if for_loop is None or structure_match[1] != for_match.group(
                    1
                ):
                    suffix = structure_match[1]
                else:
                    suffix = str(it_val)
                # change to dict-style syntax (structure.suffix -> structure["suffix"])
                line = line.replace(
                    structure_match[0],
                    "structure['{}']".format(suffix),
                )
                if suffix not in structure_dict:
                    structure_dict[suffix] = AaronTools.geometry.Geometry()
                    kind_dict[suffix] = None
            return line

        structure_dict = {}
        kind_dict = {}
        structure_list = []
        # load templates from AARONLIB
        if "Reaction" in self:
            path = None
            if "template" in self["Reaction"]:
                path = os.path.join(
                    AARONLIB,
                    "template_geoms",
                    self["Reaction"]["reaction"],
                    self["Reaction"]["template"],
                )
                for dirpath, dirnames, filenames in os.walk(path):
                    structure_list += get_multiple(filenames, path=dirpath)
            else:
                path = os.path.join(
                    AARONLIB,
                    "template_geoms",
                    self["Reaction"]["reaction"],
                )
                for dirpath, dirnames, filenames in os.walk(path):
                    structure_list += get_multiple(filenames, path=dirpath)
            for structure, kind in structure_list:
                structure.name = os.path.relpath(structure.name, path)
        if not self.has_section("Geometry"):
            return structure_list

        # load templates from config[Geometry]
        # store in structure_dict, keyed by structure option suffix
        # `structure.suffix = geom.xyz` store as {suffix: geom.xyz}
        # `structure = geom.xyz` (no suffix), store as {"": geom.xyz}
        if "structure" in self["Geometry"]:
            structure_dict[""] = self["Geometry"]["structure"]
        else:
            for key in self["Geometry"]:
                if key.startswith("structure."):
                    suffix = ".".join(key.split(".")[1:])
                    structure_dict[suffix] = self["Geometry"][key]
        # create Geometry objects
        pop_sd = set([])
        for suffix, structure in structure_dict.items():
            if structure is not None and os.path.isdir(structure):
                # if structure is a directory
                for dirpath, dirnames, filenames in os.walk(structure):
                    structure_list += get_multiple(
                        filenames, path=dirpath, suffix=suffix
                    )
            elif structure is not None:
                try:
                    # if structure is a filename
                    structure = AaronTools.geometry.Geometry(structure)
                except FileNotFoundError:
                    # if structure is a filename
                    structure = AaronTools.geometry.Geometry(
                        os.path.join(self["DEFAULT"]["top_dir"], structure)
                    )
                except (IndexError, NotImplementedError):
                    if "coordination_complex" in structure.lower():
                        shape = None
                        center = None
                        ligands = None
                        for line in structure.splitlines():
                            line = line.strip()
                            if "coordination_complex" in line.lower():
                                shape = re.split("[:=]", line)[1].strip()
                            if "center" in line.lower():
                                center = re.split("[:=]", line)[1].strip()
                            if "ligands" in line.lower():
                                ligands = (
                                    re.split("[:=]", line)[1].strip().split()
                                )
                        for (
                            geom
                        ) in AaronTools.geometry.Geometry.get_coordination_complexes(
                            center=center, ligands=ligands, shape=shape
                        )[
                            0
                        ]:
                            if suffix:
                                geom.name += "." + suffix
                            structure_list += [(geom, None)]
                        structure = None
                        pop_sd.add(suffix)
                    else:
                        # if structure is a smiles string
                        structure = AaronTools.geometry.Geometry.from_string(
                            structure
                        )
                # adjust structure attributes
                if structure is not None:
                    if "name" in self["Job"]:
                        structure.name = self["Job"]["name"]
                    if "Geometry" in self and "comment" in self["Geometry"]:
                        structure.comment = self["Geometry"]["comment"]
                        structure.parse_comment()
                    structure_dict[suffix] = structure
                    kind_dict[suffix] = None
        for s in pop_sd:
            del structure_dict[s]

        # for loop for structure modification/creation
        # structure.suffix = geom.xyz
        # &for name in <iterator>:
        #    structure.name = structure.suffix.copy()
        #    structure.name.method_call(*args, **kwargs)

        if "Geometry" in self:
            for key in self["Geometry"]:
                if not key.startswith("&for"):
                    continue
                for_match = for_patt.search(key)
                if for_match is None:
                    raise SyntaxError(
                        "Malformed &for loop specification in config"
                    )
                lines = self["Geometry"][key].split("\n")
                for it_val in eval(for_match.group(2), {}):
                    eval_dict = {
                        "Geometry": AaronTools.geometry.Geometry,
                        "structure": structure_dict,
                        for_match.group(1): it_val,
                    }
                    for line in lines:
                        line = line.strip()
                        if not line:
                            continue
                        line = structure_suffix_parse(
                            line,
                            for_loop=(for_match, it_val),
                        )
                        if "=" in line:
                            structure_assignment(line)
                        else:
                            eval(line, eval_dict)

        # add structure_dict to structure list
        try:
            padding = max(
                [
                    len(suffix)
                    for suffix in structure_dict.keys()
                    if suffix.isnumeric()
                ]
            )
        except ValueError:
            padding = 0
        for suffix in structure_dict:
            geom = structure_dict[suffix]
            if suffix:
                geom.name = "{}.{}".format(
                    self["Job"]["name"], suffix.zfill(padding)
                )
            structure_list += [(geom, kind_dict[suffix])]

        # apply functions found in [Geometry] section
        if "Geometry" in self and "&call" in self["Geometry"]:
            eval_dict = {
                "Geometry": AaronTools.geometry.Geometry,
                "structure": structure_dict,
            }
            lines = self["Geometry"]["&call"]
            for line in lines.split("\n"):
                line = structure_suffix_parse(line)
                if parsed_struct_patt.search(line.strip()):
                    try:
                        eval(line.strip(), eval_dict)
                    except SyntaxError:
                        structure_assignment(line)
                    for suffix in structure_dict:
                        val = (structure_dict[suffix], kind_dict[suffix])
                        if val not in structure_list:
                            structure_list += [val]
                elif line.strip():
                    for structure, kind in structure_list:
                        eval_dict["structure"] = structure
                        eval(line.strip(), eval_dict)
        return structure_list

    def make_changes(self, structure):
        if not self._changes:
            return structure
        changed = []
        for name, (changes, kind) in self._changes.items():
            for key, val in changes.items():
                if kind == "Substitution" and "(" not in key:
                    # regular substitutions
                    for k in key.split(","):
                        k = k.strip()
                        if val.lower() == "none":
                            structure -= structure.get_fragment(k)
                        else:
                            sub = structure.substitute(val, k)
                            for atom in sub:
                                changed += [atom.name]
                elif kind == "Substitution":
                    # fused ring substitutions
                    target_patt = re.compile("\((.*?)\)")
                    for k in target_patt.findall(key):
                        k = [i.strip() for i in k.split(",")]
                        if val.lower() == "none":
                            structure -= structure.get_fragment(*k)
                        else:
                            sub = structure.ring_substitute(k, val)
                            for atom in sub:
                                changed += [atom.name]
                elif kind == "Mapping":
                    key = [k.strip() for k in key.split(",")]
                    new_ligands = structure.map_ligand(val, key)
                    for ligand in new_ligands:
                        for atom in ligand:
                            changed += [atom.name]
        try:
            con_list = list(
                eval(
                    "[{}]".format(self["Geometry"].get("constraints", "")), {}
                )
            )
        except KeyError:
            structure.parse_comment()
            try:
                con_list = structure.other["constraint"]
            except KeyError:
                con_list = []
        for con in con_list:
            for c in con:
                try:
                    changed.remove(str(c))
                except ValueError:
                    pass
        self._changed_list = changed
        return structure

    def _parse_includes(self, section=None):
        """
        Moves option values from subsections into parent section
        Eg:
            [HPC]
            include = Wheeler
            ppn = 12

            [HPC.Wheeler]
            nodes = 1
            queue = wheeler_q

            ^^^evaluates to:
            [Job]
            nodes = 1
            ppn = 12
            queue = wheeler_q
        """
        if section is None:
            section_list = self.sections()
        elif isinstance(section, str):
            section_list = [section]
        else:
            section_list = section
        for section in section_list:
            # add requested subsections to parent section
            if self.has_option(section, "include"):
                if section == "Job" and self[section]["include"] == "detect":
                    continue
                include_section = self[section]["include"].split(".")
                if include_section[0] in self.sections():
                    # include specifies full section name, eg:
                    # include = Job.Minimum --> [Job.Minimum]
                    include_section = ".".join(include_section)
                else:
                    # short-form of include, eg:
                    # [Job]
                    # include = Minimum
                    # --> [Job.Minimum]
                    include_section = [section] + include_section
                    include_section = ".".join(include_section)
                for key, val in self[include_section].items():
                    self[section][key] = val
            # handle non-default capitalization of default section
            if section.lower() == "default":
                for key, val in self[section].items():
                    self["DEFAULT"][key] = val

    def as_dict(self, spec=None, skip=None):
        """
        Forms a metadata spec dictionary from configuration info
        :spec: (dict) if given, append key/vals to that dict
        :skip: (list) skip storing stuff according to (section, option) or attrs
               section, option, and attrs are strings that can be regex (full match only)
               eg: skip=[("Job", ".*"), "conformer"] will skip everything in the Job
               section and the Config.conformer attribute
        """
        if spec is None:
            spec = {}
        if skip is None:
            skip = []
        skip_attrs = []
        skip_sections = []
        skip_options = []
        for s in skip:
            if isinstance(s, tuple):
                skip_sections.append(s[0])
                skip_options.append(s[1])
            else:
                skip_attrs.append(s)

        for attr in self.SPEC_ATTRS:
            for s in skip_attrs:
                if re.fullmatch(s, attr):
                    break
            else:
                spec[attr] = getattr(self, attr)

        for section in ["DEFAULT"] + self.sections():
            if "." in section:
                # these are include sections that should already be pulled into
                # the main body of the config file
                continue
            for option in self[section]:
                for i, s in enumerate(skip_sections):
                    o = skip_options[i]
                    if re.fullmatch(s, section) and re.fullmatch(o, option):
                        break
                else:
                    # only include default options once, unless they are overridden
                    if (
                        section != "DEFAULT"
                        and option in self["DEFAULT"]
                        and self["DEFAULT"][option] == self[section][option]
                    ):
                        continue
                    spec[
                        "{}/{}".format(section, option.replace(".", "_"))
                    ] = self[section][option]
        return spec

    def read_spec(self, spec):
        """
        Loads configuration metadata from spec dictionaries
        """
        for attr in spec:
            if attr in self.SPEC_ATTRS:
                setattr(self, attr, spec[attr])
            if "/" in attr:
                section, key = attr.split("/")
                if section not in self:
                    self.add_section(section)
                self[section][key] = spec[attr]

    def for_step(self, step=None, parse_functions=True):
        """
        Generates a config copy with only options for the given step
        """
        config = self.copy()
        # find step-specific options
        for section in ["DEFAULT"] + config.sections():
            for key, val in config[section].items():
                remove_key = key
                key = key.strip().split()
                if len(key) == 1:
                    continue
                try:
                    key_step = float(key[0])
                except ValueError:
                    continue
                key = " ".join(key[1:])
                # screen based on step
                if key_step == float(step):
                    config[section][key] = val
                # clean up metadata
                del config[section][remove_key]
        # other job-specific additions
        if config.has_section("HPC") and "host" in config["HPC"]:
            try:
                config["HPC"]["work_dir"] = config["HPC"].get("remote_dir")
            except TypeError as e:
                raise RuntimeError(
                    "Must specify remote working directory for HPC (remote_dir = /path/to/HPC/work/dir)"
                ) from e
        else:
            if not config.has_section("HPC"):
                config.add_section("HPC")
            config["HPC"]["work_dir"] = config["DEFAULT"].get("top_dir")
        # parse user-supplied functions in config file
        if parse_functions:
            config.parse_functions()
        return config
